#include <cstdlib>
#include <cstring>
#include <vector>
#include <map>
#include <zlib.h>
#include <sys/stat.h>
#include "bgzf.h"
#include "kstring.h"
#include "stats.cpp"
#include <cassert>
const char * BIN= ".counts.bin";
const char * IDX= ".counts.idx";
//does a file exists?
int fexists(const char* str){///@param str Filename given as a string.
  struct stat buffer ;
  return (stat(str, &buffer )==0 ); /// @return Function returns 1 if file exists.
}


struct ltstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) < 0;
  }
};

typedef struct{
  int nSites;
  int64_t fpos;
}datum;

typedef std::map<char*,datum,ltstr> myMap;



//return one+two
char *append(char *one,const char*two){
  char *ret = new char[strlen(one)+strlen(two)+1];
  strcpy(ret,one);
  strcpy(ret+strlen(ret),two);
  return ret;
}


#define LENS  4096

void writemap(FILE *fp,const myMap &mm){
  fprintf(fp,"\t\tInformation from index file:\n");
  int i=0;
  for(myMap::const_iterator it=mm.begin();it!=mm.end();++it){
    datum d = it->second;
    fprintf(fp,"\t%d\t%s\t%d\t%ld\n",i++,it->first,d.nSites,(long int)d.fpos);

  }

}


myMap getMap(const char *fname){
  myMap ret;
  int clen;
  if(!fexists(fname)){
    fprintf(stderr,"Problem opening file: %s\n",fname);
    exit(0);
  }
  FILE *fp = fopen(fname,"r");
  while(fread(&clen,sizeof(int),1,fp)){
  
    char *chr = new char[clen+1];
    int hasRead =fread(chr,1,clen,fp); 
  
    if(clen!=hasRead){
      fprintf(stderr,"Problem with reading chr: clen:%d hasread:%d\n",clen,hasRead);
      exit(0);
    }
    chr[clen] = '\0';
    //    fprintf(stderr,"chr is:%s\n",chr);
    datum d;
    if(1!=fread(&d.nSites,sizeof(int),1,fp)){
      fprintf(stderr,"[%s.%s():%d] Problem reading data: %s \n",__FILE__,__FUNCTION__,__LINE__,fname);
      exit(0);
    }
    //fprintf(stderr,"len:%d\n",d.nSites);
    if(1!=fread(&d.fpos,sizeof(int64_t),1,fp)){
      fprintf(stderr,"[%s.%s():%d] Problem reading data: %s \n",__FILE__,__FUNCTION__,__LINE__,fname);
      exit(0);
    }
    myMap::iterator it = ret.find(chr);
    if(it==ret.end())
      ret[chr] =d ;
    else{
      fprintf(stderr,"Problem with chr: %s, key already exists\n",chr);
      exit(0);
    }
  }

  return ret;
}

typedef struct{
  char *chr;
  int l;
  int m;
  int *posi;
  char **counts;
}perChr;

void dalloc(perChr &pc){
  delete [] pc.chr;
  delete [] pc.posi;
  for(int i=0;i<4;i++)
    delete [] pc.counts[i];
  delete [] pc.counts;
    
}



void getPerChr(BGZF *fp,perChr &ret){
  
  int clen;
  if(bgzf_read(fp,&clen,sizeof(int))==0){
    ret.l =0;
    return;
  }
  ret.chr = new char[clen+1];
  bgzf_read(fp,ret.chr,clen);
  ret.chr[clen] = '\0';
  int newsize;
  bgzf_read(fp,&newsize,sizeof(int));
  if(newsize>ret.m){
    for(int i=0;i<4;i++){
      delete [] ret.counts[i];
      ret.counts[i] = new char[newsize];
    }
    ret.m=newsize;
  }
  ret.l=newsize;
  for(int i=0;i<4;i++)
    bgzf_read(fp,ret.counts[i],ret.l*sizeof(char));  
 
}



void print_main(perChr &pc,FILE *fp){
  for(int i=0;i<pc.l;i++)
    fprintf(fp,"%s\t%d\t%d\t%d\t%d\t%d\n",pc.chr,i+1,pc.counts[0][i],pc.counts[1][i],pc.counts[2][i],pc.counts[3][i]);
}

int print(int argc, char**argv){
  if(argc==0){
    fprintf(stderr,"print FILE [-r chrName]\n");
    exit(0);
  }
  char *base = *argv;
  char* outnames_bin = append(base,BIN);
  char* outnames_idx = append(base,IDX);
  fprintf(stderr,"Assuming binfile:%s and indexfile:%s\n",outnames_bin,outnames_idx);
  
  myMap mm = getMap(outnames_idx);
  writemap(stderr,mm);
  BGZF *fp = bgzf_open(outnames_bin,"r");

  --argc;++argv;
  //  fprintf(stderr,"argc=%d\n",argc);
  int argP =0;
  char *chr=NULL;

  while(argP<argc){
    //   fprintf(stderr,"args=%s\n",argv[argP]);
    if(argP==argc){
      fprintf(stderr,"incomplete arguments list\n");
      exit(0);
    }
    if(strcmp("-r",argv[argP])==0)
      chr = argv[argP+1];
    else {
      fprintf(stderr,"Unknown argument:%s\n",argv[argP]);
      exit(0);

    }
    argP +=2;
  }
  
  
  if(chr!=NULL){  
    myMap::iterator it = mm.find(chr);
    if(it==mm.end()){
      fprintf(stderr,"Problem finding chr: %s in index\n",chr);
      exit(0);
    }
    datum d = it->second;
    bgzf_seek(fp,d.fpos,SEEK_SET);
  }
  perChr pc;
  pc.counts = new char*[4];
  for(int i=0;i<4;i++) pc.counts[i] = NULL;
  pc.chr=NULL;
  pc.posi=NULL;
  pc.l=pc.m=0;
  while(1){
    getPerChr(fp,pc);
    if(pc.l==0)
      break;
    //    fprintf(stderr,"pc.chr=%s pc.nSites=%d firstpos=%d lastpos=%d\n",pc.chr,pc.nSites,pc.posi[0],pc.posi[pc.nSites-1]);
    print_main(pc,stdout);
    if(chr!=NULL)
      break;
    dalloc(pc);
  }

  return 0;
}
int extract(int argc, char**argv,FILE *fpout) {
  if(argc!=2){
    fprintf(stderr,"print FILE.bam FILE.bim \n");
    exit(0);
  }
  char *base = *argv;
  char* outnames_bin = append(base,BIN);
  char* outnames_idx = append(base,IDX);
  fprintf(stderr,"Assuming binfile:%s and indexfile:%s\n",outnames_bin,outnames_idx);
  
  myMap mm = getMap(outnames_idx);
  writemap(stderr,mm);
  BGZF *fp = bgzf_open(outnames_bin,"r");

  --argc;++argv;
  
  //  fprintf(stderr,"adfasdf:%s\n",*argv);
  
  gzFile gz=Z_NULL;
  gz=gzopen(*argv,"r");
  if(gz==Z_NULL){
    fprintf(stderr,"Problem opening file:%s\n",*argv);
    exit(0);
  }

  perChr pc;
  pc.counts = new char*[4];
  for(int i=0;i<4;i++) pc.counts[i] = NULL;
  pc.posi=NULL;
  pc.l=pc.m=0;


  char *buf = new char [LENS];
  
  char *last =NULL;
  while(gzgets(gz,buf,LENS)){
    char *chr = strtok(buf,"\n\t ");
    if(last==NULL||strcmp(chr,last)!=0){
      //load chr
      myMap::iterator it = mm.find(chr);
      if(it==mm.end()){
	fprintf(stderr,"Problem finding chr: %s in index\n",chr);
	exit(0);
      }
      datum d = it->second;
      bgzf_seek(fp,d.fpos,SEEK_SET);
      getPerChr(fp,pc);
      last=strdup(chr);//<- small leak, only happens when we change chr
    }
    int p=atoi(strtok(NULL,"\t\n "))-1;
    if(p>pc.l){
      fprintf(stderr,"position to extract is after end of chromosome:\n\n");fflush(stderr);
      exit(0);
    }

    fprintf(fpout,"%s\t%d",pc.chr,p+1);
    for(int i=0;i<4;i++)
      fprintf(fpout,"\t%d",(int)pc.counts[i][p]);
    fprintf(fpout,"\n");
  }

  dalloc(pc);
  return 0;
}

int main(int argc,char **argv){
  if(argc==1){
    fprintf(stderr,"./smartCount [print extract]\n");
    return 0;
  }
  //  fprintf(stderr,"zlibversion=%s zlibversion=%s file:%s\n",ZLIB_VERSION,zlib_version,__FILE__);
  --argc,++argv;
  if(strcmp(*argv,"print")==0){
    print(--argc,++argv);
  }else if(strcmp(*argv,"extract")==0){
    extract(--argc,++argv,stdout);
  }else{
    fprintf(stderr,"Unknown argument: \'%s' please supply:\n",*argv);
    fprintf(stderr,"./smartCount [print|extract]\n");
  }
}

